#!/usr/bin/env python

from __future__ import print_function

import argparse
from datetime import datetime, timedelta
import hashlib
from httplib import HTTPSConnection
from lxml import etree
import re
from suds import WebFault
from suds.client import Client
from suds.plugin import MessagePlugin
from suds.transport.http import HttpTransport
import sys
from urllib2 import build_opener, HTTPSHandler, URLError

__version__ = "2.1"

REGEX = r'CN=(?P<hash>[0-9a-f]{8}), CN=(?P<date>[0-9]{8}), CN=Test Subject, CN=000001, CN=tsubject, OU=Users, OU=Organic Units, DC=cern, DC=ch'
XML_NS = {
  'soapenv': 'http://schemas.xmlsoap.org/soap/envelope/',
  'pap': 'http://services.pap.authz.glite.org',
  'xacml': 'urn:oasis:names:tc:xacml:2.0:policy:schema:os',
}
XPATH_LISTPOLICIES = '/soapenv:Envelope/soapenv:Body/pap:listPoliciesResponse/pap:listPoliciesReturn/xacml:Policy/xacml:Rule/xacml:Target/xacml:Subjects/xacml:Subject/xacml:SubjectMatch/xacml:AttributeValue'
XPATH_GETALLPAPS = '/soapenv:Envelope/soapenv:Body/pap:getAllPapsResponse/pap:getAllPapsReturn/pap:getAllPapsReturn/pap:alias'


NAGIOS_OK = 0
NAGIOS_WARNING = 1
NAGIOS_CRITICAL = 2
NAGIOS_UNKNOWN = 3


class MyHTTPSHandler(HTTPSHandler):
    def __init__(self, key, cert):
        HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert

    def https_open(self, req):
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=5):
        return HTTPSConnection(host, key_file=self.key, cert_file=self.cert, timeout=timeout)


class MyHTTPSTransport(HttpTransport):
    def __init__(self, key, cert, *args, **kwargs):
        HttpTransport.__init__(self, *args, **kwargs)
        self.key = key
        self.cert = cert

    def u2open(self, u2request):
        timeout = self.options.timeout
        url = build_opener(MyHTTPSHandler(self.key, self.cert))
        return url.open(u2request, timeout=timeout)


def validate_DN(dn, salt):
    data = re.match(REGEX, dn)
    if data is None:
        return
    sha = hashlib.new('sha256')
    sha.update(salt)
    sha.update(data.group('date'))
    sig = sha.hexdigest()
    if sig[0:8] == data.group('hash'):
        return data.group('date')


class ListPoliciesFilter(MessagePlugin):
    def __init__(self, salt, result, raw=False, all_raw=False):
        self._salt = salt
        self._result = result
        self._raw = raw or all_raw
        self._all_raw = all_raw

    def received(self, context):
        text = context.reply
        context.reply = ''
        root = etree.XML(text)
        for node in root.xpath(XPATH_LISTPOLICIES, namespaces=XML_NS):
            if node.attrib['DataType'] == 'urn:oasis:names:tc:xacml:1.0:data-type:x500Name':
                res = validate_DN(node.text, self._salt)
                if res is not None:
                    self._result.add(res)
                    if self._raw:
                        print(node.text, file=sys.stderr)
                elif self._all_raw:
                    print(node.text, file=sys.stderr)


class GetAllPapsFilter(MessagePlugin):
    def __init__(self, result):
        self._result = result

    def received(self, context):
        text = context.reply
        context.reply = ''
        root = etree.XML(text)
        for node in root.xpath(XPATH_GETALLPAPS, namespaces=XML_NS):
            self._result.add(node.text)


def get_dyn_dn_dates(server, cert, key, salt, raw=False, all_raw=False, timeout=9):
    aliases = set()
    client = Client('https://{0}:8150/pap/services/PAPManagementService?wsdl'.format(server),
                    transport=MyHTTPSTransport(key, cert, timeout=timeout/2.),
                    plugins=[GetAllPapsFilter(aliases)])
    client.service.getAllPaps()
    res = set()
    client = Client('https://{0}:8150/pap/services/XACMLPolicyManagementService?wsdl'.format(server),
                    transport=MyHTTPSTransport(key, cert, timeout=timeout/2.),
                    plugins=[ListPoliciesFilter(salt, res, raw=raw, all_raw=all_raw)])
    for alias in aliases:
        client.service.listPolicies(alias)
    return res


def parse_args():
    parser = argparse.ArgumentParser(description='Extract and verify daily dynamic DNs from Argus')
    parser.add_argument('-V', '--version', action='version', version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument("-v", "--verbosity", action="count", help="increase output verbosity")
    parser.add_argument("-t", "--timeout", action="store", type=int, help="Set timeout for the probe", default=9)
    parser.add_argument('--cert', required=True, help='Certificate used for the connection')
    parser.add_argument('--key', required=True, help='Private key used for the connection')
    parser.add_argument('--salt', required=True, help='Salt for the dyn dn')
    parser.add_argument('-H', help='Server to connect to', default='lcg-argus.cern.ch')
    parser.add_argument('--valid', help='Valid dates', action='append')
    parser.add_argument('--today', help='Check only for today', action='store_true')
    parser.add_argument('--raw', help='Get dynamic DNs on stderr', action='store_true')
    parser.add_argument('--all-raw', help='Get all DNs on stderr', action='store_true', dest='all_raw')
    return parser.parse_args()


def print_dns_validity(verbosity, dns, ok, warn):
    if verbosity < 2:
        return
    for dn in dns:
        if dn in ok:
            print('+{0}'.format(dn))
        elif dn in warn:
            print('-{0}'.format(dn))
        else:
            print('?{0}'.format(dn))


def main(args):
    try:
        dns = get_dyn_dn_dates(args.H, args.cert, args.key, args.salt, raw=args.raw, all_raw=args.all_raw, timeout=args.timeout)
    except URLError, e:
        print('ARGUS WARN - connection error')
        if args.verbosity >= 2:
            print(e.args)
        return NAGIOS_WARNING
    except WebFault, e:
        print('ARGUS WARN - Authorization error')
        if args.verbosity >= 2:
            print(e.args)
        return NAGIOS_WARNING
    dns = sorted(dns)
    if args.valid is None and not args.today:
        print('ARGUS OK - No validity required')
        print_dns_validity(args.verbosity, dns, [], [])
        return NAGIOS_OK
    if args.today:
        today = datetime.now().strftime("%Y%m%d")
        yesterday = (datetime.now() - timedelta(days=1)).strftime("%Y%m%d")
        if today in dns:
            print('ARGUS OK - Found expected DN')
            print_dns_validity(args.verbosity, dns, [today], [yesterday])
            return NAGIOS_OK
        elif yesterday in dns:
            print('ARGUS WARN - Found outdated DN')
            print_dns_validity(args.verbosity, dns, [today], [yesterday])
            return NAGIOS_WARNING
        else:
            print('ARGUS CRIT - Expected DN not found!')
            print_dns_validity(args.verbosity, dns, [today], [yesterday])
            return NAGIOS_CRITICAL
    elif args.valid:
        ok = set()
        for valid in args.valid:
            if valid in dns:
                ok.add(valid)
        if ok:
            print('ARGUS OK - Found expected DN')
            print_dns_validity(args.verbosity, dns, valid, [])
            return NAGIOS_OK
        else:
            print('ARGUS CRIT - Expected DN not found!')
            print_dns_validity(args.verbosity, dns, valid, [])
            return NAGIOS_CRITICAL
    print('ARGUS CRIT - broken probe (this path should not be taken)')
    print_dns_validity(args.verbosity, dns, [], [])
    return NAGIOS_CRITICAL


if __name__ == '__main__':
    sys.exit(main(parse_args()))
